unit MsgSock;

{******************************************************************************
 Nome:			MsgSock
 Scopo:     MsgSock permette lo scambio di dati tra diversi processi utiliz-
            zando messaggi sia sincroni che asincroni.
 Generalit:MsgSock contiene il componente TMessageSocket che permette di
 						scambiare dati tra diversi processi WIN32 che ospitino una sua
            istanza utilizzando messaggi. Lo scambio pu essere sincrono
            (equivalente ad una SendMessage) o asincrono (equivalente ad una
            PostMessage).
 Autore:    Ing. M. Venturini
 Revisore:
 Data:      22 Dec. 97
 Revisioni:
******************************************************************************}

interface

uses
  SysUtils, Windows, Messages, Classes, Graphics, Controls, Forms;

{******************************************************************************
 -- Property - published
    PostSupport:		read/write Boolean
    								Il valore true abilita la gestione degli scambi di messaggi
                    asincroni, in caso sono gestiti solo quelli sincroni.
    LowMessageID:   read/write Word
    HighMessageID:  read/write Word
    								Vengono gestiti solo i messaggi con identificatori tra
                    LowMessageID e HighMessageID. Il valore LowMessageID 
                    riservato a BroadcastIdentity e non deve essere usato con
                    SendMessage e PostMessage.
    OnSentMessage:  read/write TMessageEvent
    								Event handler per l'evento SentMessage.
    OnPostedMessage:read/write TMessageEvent
    								Event handler per l'evento PostedMessage.
******************************************************************************}

{******************************************************************************
 --	Event SentMessage
 E' attivato a seguito della presenza di un messaggio inviato con SendMessage.
 La thread entro cui  stata chiamata SendMessage  sospesa fino al termine
 dell'handler o alla chiamata a MessageHandled.
 -- Parametri
    Sender: 				in TObject
                    Ogggetto attivante l'evento.
    MessageID:			in Word
                    Identificatore del messaggio.
    SourceHandle:   in Hwnd
                    Handle del TMessageSocket tramite cui  stato inviato il
                    messaggio. Pu essere utilizzato per inviare una risposta.
    DataPtr:  			in Pointer
    								Puntatore ai dati del messaggio.
    DataSize:				in Word
                    Size dei dati del messaggio.
******************************************************************************}

{******************************************************************************
 --	Event PostedMessage
 E' attivato a seguito della presenza di un messaggio inviato con PostMessage.
 La thread entro cui  stata chiamata PostMessage non  sospesa.
 -- Parametri
    Sender: 				in TObject
                    Ogggetto attivante l'evento.
    MessageID: 			in Word
                    Identificatore del messaggio.
    SourceHandle:   in Hwnd
                    Handle del TMessageSocket tramite cui  stato inviato il
                    messaggio. Pu essere utilizzato per inviare una risposta.
    DataPtr:  			in Pointer
    								Puntatore ai dati del messaggio.
    DataSize:				in Word
                    Size dei dati del messaggio.
******************************************************************************}

{******************************************************************************
 --	method PostSupported
 Permette di verificare se un componente TMessageSocket cui si vuole inviare un
 messaggio  in grado di gestire scambi di messaggi asincroni.
 -- Parametri
    Handle:         in Hwnd
    								Handle ottenuto da un evento SentMessage o PostedMessage.
		Result:					out Boolean
    								Se true il componente corrispondente all'handle pu gestire
                    scambi di messaggi asincroni.
******************************************************************************}

{******************************************************************************
 --	method BroadcastIdentity
 Tramite BroadcastIdentity si notifica la presenza di un componente
 TMessageSocket utilizzando un messaggio di broadcast. Questo provoca da parte
 dei componenti entro altri processi che hanno lo stesso valore di LowMessageID
 l'invio di messaggi cui corrisponde l'attivazione di SentMessage. Nell'handler
 MessageID  pari a LowMessageID, SourceHandle  l'handler del componente da
 cui proviene la risposta, DataPtr  nil e DataSize  0.
******************************************************************************}

{******************************************************************************
 --	method SendMessage
 SendMessage permette di inviare un messaggio a un componente TMessageSocket.
 L'invio  sincrono, quindi SendMessage termina dolo dopo il completamento
 dell'esecuzione dell'evento SentMessage del componente di destinazione o dopo
 che entro di esso sia stato chiamato MessageHandled.
 -- Parametri
		MessageID:			in Word
    								Identificatore per il messaggio. Deve essere maggiore di
                    LowMessageID e minore o uguale a HighMessageID.
    DestinationHandle: in Hwnd
    								Handle del componente TMessageSocket a cui si vuole inviare
                    il messaggio.
    DataPtr:				in Pointer
    								Puntatore ai dati che si vogliono trasmettere con il
                    messaggio.
    DataSize:       in Word
    								Size dei dati che si vogliono trasmettere con il messaggio.
******************************************************************************}

{******************************************************************************
 --	method PostdMessage
 PostMessage permette di inviare un messaggio a un componente TMessageSocket.
 L'invio  asincrono, quindi PostMessage termina prima del completamento
 dell'esecuzione dell'evento PostMessage del componente di destinazione.
 -- Parametri
		MessageID:			in Word
    								Identificatore per il messaggio. Deve essere maggiore di
                    LowMessageID e minore o uguale a HighMessageID.
    DestinationHandle: in Hwnd
    								Handle del componente TMessageSocket a cui si vuole inviare
                    il messaggio.
    DataPtr:				in Pointer
    								Puntatore ai dati che si vogliono trasmettere con il
                    messaggio.
    DataSize:       in Word
    								Size dei dati che si vogliono trasmettere con il messaggio.
******************************************************************************}

{******************************************************************************
 --	method MessageHandled
 Chiamata entro un handler dell'evento SentMessage segnala il completamento
 dell'elaborazione del messaggio e sblocca l'esecuzione della thread che ha
 inviato il messaggio. Dopo la sua chiamata i dati puntati dal parametro
 DataPtr dell'evento non sono pi validi.
******************************************************************************}

type
  TMessageEvent = procedure(			Sender: TObject;
  				 						  		const	MessageID: Word;
  													const	SourceHandle: Hwnd;
    												const	DataPtr: Pointer;
                        		const	DataSize: Word) of object;
  TMessageSocket = class(TComponent)
  private
  	FPostThreadPriority: TThreadPriority;
  	FPostSupport: Boolean;
    FLowMessageID: Word;
    FHighMessageID: Word;
    FOnSentMessage,FOnPostedMessage: TMessageEvent;
    PostThread: TThread;
    SendWindow: Hwnd; {Handle della window associata al componente,  usata per identificare il componente stesso}
    PostWindow: Hwnd; {Handle della window associata alla PostThread}
    HandlesList: Pointer;
    procedure SendWindowProc(var	Msg: TMessage);
    procedure SetPostThreadPriority(Priority: TThreadPriority);
    procedure SetLowMessageID(ID: Word);
    procedure SetHighMessageID(ID: Word);
    function PostHandle(const SendHandle: Hwnd): Hwnd;
    procedure Loaded; override;
  public
	  function PostSupported(const Handle: Hwnd): Boolean;
		procedure BroadcastIdentity;
   	procedure SendMessage(const	MessageID: Word;
    											const DestinationHandle: Hwnd;
    											const DataPtr: Pointer;
                      		const DataSize: Word);
   	procedure PostMessage(const	MessageID: Word;
    											const DestinationHandle: Hwnd;
    											const DataPtr: Pointer;
                      		const DataSize: Word);
    procedure MessageHandled;
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
  	property PostThreadPriority: TThreadPriority read FPostThreadPriority write SetPostThreadPriority;
  	property PostSupport: Boolean read FPostSupport write FPostSupport;
  	property LowMessageID: Word read FLowMessageID write SetLowMessageID;
  	property HighMessageID: Word read FHighMessageID write SetHighMessageID;
  	property OnSentMessage: TMessageEvent read FOnSentMessage write FOnSentMessage;
  	property OnPostedMessage: TMessageEvent read FOnPostedMessage write FOnPostedMessage;
  end;
  EMessageSocket = class(Exception);
  EIllegalID = class(EMessageSocket);
  EPostHandle = class(EMessageSocket);
  EPostUnsupported = class(EMessageSocket);
  EPostThreadPriority = class(EMessageSocket);

procedure Register;

{* Implementation ************************************************************}

implementation

const
	EIllegalIDMsg = ' - Illegal ID: ';
  EPostHandleMsg = ' - Post Handle not found';
  EPostUnsupportedMsg = ' - Destination does not support Posts';
  EPostThreadPriorityMsg = ' - Priority must be tpHigher or above';	

type
  {Record per la lista con gli handler dei componenti che rispondono alla BroadcastIdentity}
 	PHandles = ^THandles;
	THandles = record
  	Send, Post: Hwnd;
    Next: PHandles;
  end;
	{Record da usare per copiare i dati dalla send queue di PostThread alla post queue della main thread}
  TDataCarrier = record
  	DataSize: Word;
    Data: array	[0..0] of Byte;
  end;
  {PostThread, riceve i WM_Copydata sulla sua send queue, ne estrae i dati,
   li copia in uno user message che invia alla post queue della main thread}
	TPostThread = class(TThread)
  private
    FLowMessageID: Word;
    FHighMessageID: Word;
    SendWindow: Hwnd;
    PostWindow: Hwnd;
    CreationEvent: THandle;
    procedure PostWindowProc(var	Msg: TMessage);
  public
  	property LowMessageID: Word read FLowMessageID write FLowMessageID;
  	property HighMessageID: Word read FHighMessageID write FHighMessageID;
  	procedure Execute; override;
    constructor Create(const 	Priority: TThreadPriority;
    									 const	SendWindow: Hwnd;
                       var		PostWindow: Hwnd);
  end;

{* TPostThread ***************************************************************}

{--	procedure PostWindowProc
 Window procedure associata alla PostThread che gestisce i messaggi di tipo
 asincrono. I dati dai messaggi WM_Copydata sincroni sono prelevati e posti
 in messaggi WM_User inviati poi alla main thread. Le thread che hanno inviato
 i WM_Copydata sono quindi immediatamente sbloccate.}

procedure TPostThread.PostWindowProc(var	Msg: TMessage);

var
  CopyData: ^TCopyDataStruct;
  DataCarrier: ^TDataCarrier;

begin

  with Msg do
  	if Msg = WM_COPYDATA then begin  {Su WM_Copydata...}
     	CopyData := Pointer(LParam);
      with CopyData^ do begin
      	if (dwData > (WM_User + FLowMessageID)) and (dwData <= (WM_User + FHighMessageID)) then begin {...se il dati sono destinati a questo componente...}
		      try
						{...si copiano i dati in DataCarrier...}
    		    GetMem(DataCarrier,SizeOf(DataCarrier^) - SizeOf(DataCarrier^.Data) + cbData);
        		if Pointer(lpData) = nil then
							DataCarrier^.DataSize := 0
    		    else begin
        		  DataCarrier^.DataSize := cbData;
		 	        if cbData > 0 then
   			        Move((Pointer(lpData))^,DataCarrier^.Data,cbData);
        		end;
		     		Result := Longint(True);
						{...e si invia il DataCarrier nella coda di post della main thread, WParam, con
             l'handle del componente da cui proviene il messaggio,  trasferito invariato  }
    		   	Windows.SendNotifyMessage(SendWindow,dwData,WParam,Longint(DataCarrier));
		      except
    		 		Result := Longint(True);
       			if Windows.InSendMessage then
		 	       	Windows.ReplyMessage(Longint(True));
   			   	Application.HandleException(Self);
		      end;
    	  end
        else
		      Result := DefWindowProc(PostWindow,Msg,WParam,LParam);
    	end;
    end
    else if Msg = WM_Quit then  {Su WM_Quit occoorre terminare PostWindowProc e tornare ad Execute}
    	Windows.PostQuitMessage(WParam)
    else
      Result := DefWindowProc(PostWindow,Msg,WParam,LParam);

end;

{-- procedure Execute
 Corpo della PostThread.}

procedure TPostThread.Execute;

var
	Event: Longint;
  Msg: TMsg;

begin

	{Si crea PostWindow e si attiva CreationEvent per riattivare la main thread}
  PostWindow := AllocateHWnd(PostWindowProc);
  Windows.SetEvent(CreationEvent);
	{Al primo messaggio si trasferisce il controllo a PostWindowProc...}
 	Windows.WaitMessage;
  while Windows.GetMessage(Msg,PostWindow,0,0) do begin
		Windows.TranslateMessage(Msg);
  	Windows.DispatchMessage(Msg);
  end;
  {... che lo restituisce solo a seguito della PostQuitMessage}
  DeallocateHWnd(PostWindow);

end;

{-- constructor Create
 Crea la PostThread e coordina lo scambio degli handle delle window.}

constructor TPostThread.Create(const	Priority: TThreadPriority;
    									 				 const	SendWindow: Hwnd;
      					               var		PostWindow: Hwnd);
begin

	inherited Create(true);
  {Per garantire che PostThread non sia bloccata dalla main thread occorre alzarne la priorit}
  Self.Priority := Priority;
  Self.SendWindow := SendWindow;
  {Resume dell'esecuzione di PostThread ed attesa su CreationEvent che la sua l'inizializzazione sia completa}
 	CreationEvent := CreateEvent(nil,Longbool(false),Longbool(false),nil);
  Resume;
  Windows.WaitForSingleObject(CreationEvent,Infinite);
  Windows.CloseHandle(CreationEvent);
  PostWindow := Self.PostWindow;  {Valido a seguito dell'attesa su CreationEvent}

end;

{* TMessageSock **************************************************************}

{--	procedure SendWindowProc
 Window procedure della window associata al componente. Essa gestisce
 direttamente i messaggi sincroni e indirettamente, tramite PostThread,
 quelli asincroni attivando i corrispondenti eventi. Essa provvede inoltre
 direttamente alla risposta al messaggio inviato dalla BroadcastIdentity. }

procedure TMessageSocket.SendWindowProc(var	Msg: TMessage);

var
  CopyData: ^TCopyDataStruct;
  DataCarrier: ^TDataCarrier;
	Handles: PHandles;

begin

  with Msg do
    if Msg = WM_COPYDATA then begin {WM_Copydat}
    	{Si attiva l'evento SentMessage estrando i parametri dal messaggio}
     	CopyData := Pointer(LParam);
      with CopyData^ do begin
        if (dwData > (WM_User + FLowMessageID)) and (dwData <= (WM_User + FHighMessageID)) and Assigned(FOnSentMessage) then begin
		  		try
						FOnSentMessage(Self,dwData - WM_User,Hwnd(WParam),Pointer(lpData),cbData);
    		    Result := Longint(Longint(True));
      		except
       			Result := Longint(True);
	     			if Windows.InSendMessage then
  	     			Windows.ReplyMessage(Longint(True));
		       	Application.HandleException(Self);
  		   	end
      	end
      	else
	 		  	Result := DefWindowProc(SendWindow,Msg,WParam,LParam);
      end;
    end
    else if Msg = (WM_User + FLowMessageID) then begin {Messaggio di BroadcastIdentity}
      if LParam = 0 then begin {Messaggio di richiesta}
      	if Hwnd(WParam) <> SendWindow then begin {Controllo che il messaggio provenga da un altro componente}
		      Result := Longint(True);
          {Si invia un messaggio (asincrono per evitare deadlock) di risposta con il valore degli handle}
          Windows.PostMessage(Hwnd(WParam),WM_User + FLowMessageID,SendWindow,Longint(PostWindow));
        end
        else
		      Result := DefWindowProc(SendWindow,Msg,WParam,LParam);
      end
      else begin {Messaggio di risposta}
      	{Si crea un record per gli handle e lo si appende alla lista...}
       	GetMem(Handles,SizeOf(THandles));
        Handles^.Send := Hwnd(WParam);
        Handles^.Post := Hwnd(LParam);
        Handles^.Next := HandlesList;
        HandlesList := Handles;
        {...e si attiva l'evento SentMessage}
        if Assigned(FOnSentMessage) then
					FOnSentMessage(Self,FLowMessageID,Hwnd(WParam),nil,0);
      end;
    end
    else if (Msg > (WM_User + FLowMessageID)) and (Msg <= (WM_User + FHighMessageID)) and Assigned(FOnPostedMessage) then begin
 	  	try
        DataCarrier := Pointer(LParam);
        FOnPostedMessage(Self,Msg - WM_User,Hwnd(WParam),@(DataCarrier^.Data),DataCarrier^.DataSize);
        Result := Longint(True);
        FreeMem(DataCarrier);
			except
       	Result := Longint(True);
       	Application.HandleException(Self);
      end;
    end
    else
      Result := DefWindowProc(SendWindow,Msg,WParam,LParam);

end;

{-- procedure SetPostThreadPriority
 Tramite essa si blocca l'assegnamento di valori illegali a PostThreadPriority}

procedure TMessageSocket.SetPostThreadPriority(Priority: TThreadPriority);
begin

	if Priority >= tpHigher then
    FPostThreadPriority := Priority
	else
  	raise EPostThreadPriority.Create( TMessageSocket.ClassName + EPostThreadPriorityMsg);

end;

{-- procedure SetLowMessageID/SetHighMessageID
 Tramite esse si blocca l'assegnamento di valori illegali a LowMessageID e HighMessageID.}

procedure TMessageSocket.SetLowMessageID(ID: Word);
begin

	if ID > ($7FFF - WM_User) then
  	raise EIllegalID.Create( TMessageSocket.ClassName + EIllegalIDMsg + IntToStr(ID))
  else begin
  	FLowMessageID := ID;
    if ID > FHighMessageID then
    	FHighMessageID := ID;
  end;

end;

procedure TMessageSocket.SetHighMessageID(ID: Word);
begin

	if ID > ($7FFF - WM_User) then
  	raise EIllegalID.Create( TMessageSocket.ClassName + EIllegalIDMsg + IntToStr(ID))
  else begin
    if ID < FLowMessageID then
    	ID := FLowMessageID;
  	FHighMessageID := ID;
  end;

end;

{-- function PostHandle
 Dato l'handle di un componente TMessageSocket restituisce il corrispondente
 handle della PostWindow cui devono essere inviati i messaggi asincroni.    }

function TMessageSocket.PostHandle(const SendHandle: Hwnd): Hwnd;

var
	Handles: PHandles;

begin

	Result := 0;
	Handles := HandlesList;
  while Handles <> nil do begin
  	if Handles^.Send = SendHandle then begin
    	Result := Handles^.Post;
    	if Result <> SendHandle then
	      Exit
      else
			 	raise EPostUnsupported.Create( TMessageSocket.ClassName + EPostUnsupportedMsg);
    end;
    Handles := Handles^.Next;
  end;
 	raise EPostHandle.Create( TMessageSocket.ClassName + EPostHandleMsg);

end;

{-- procedure Loaded
 Al caricamento del componente si crea, se richiesto, la PostThread.}

procedure TMessageSocket.Loaded;
begin

	{A Design time la PostThread non va creata.}
	if not (csDesigning in ComponentState) then begin
   	if FPostSupport then begin
	 		{La PostThread va creata solo se  richiesto il supporto dei messaggi asincroni}
		  PostThread := TPostThread.Create(FPostThreadPriority,SendWindow,PostWindow);
			TPostThread(PostThread).LowMessageID := FLowMessageID;
			TPostThread(PostThread).HighMessageID := FHighMessageID;
    end
    else
			{L'uguaglianza dei due handle segnala l'assenza di supporto dei messaggi asincroni}
    	PostWindow := SendWindow;
  end;

end;

{-- constructor Create
 Con Create si crea il componente e la associata SendWindow.}

constructor TMessageSocket.Create(AOwner: TComponent);
begin

	inherited Create(AOwner);
  FPostThreadPriority := tpHigher;
  FPostSupport := false;
  SendWindow := AllocateHWnd(SendWindowProc);
  PostWindow := 0;
	HandlesList := nil;

end;

{-- destructor Destroy
 Con Destroy si distrugge il componente, la associata SendWindow, l'eventuale PostThread
 e la lista con gli handler dei componenti da cui sono arrivate risposte alla BroadcastIdentity.}

destructor TMessageSocket.Destroy;

var
	Handles: PHandles;

begin

	{A Design time la PostThread non viene creata e quindi non va distrutta}
	if not (csDesigning in ComponentState) and FPostSupport then begin
	  Windows.SendMessage(PostWindow,WM_Quit,0,0);
		PostThread.WaitFor;
  	PostThread.Free;
  end;
  DeallocateHWnd(SendWindow);
	while HandlesList <> nil do begin
		Handles := HandlesList;
    HandlesList := Handles^.Next;
    FreeMem(Handles);
  end;
  inherited Destroy;

end;

function TMessageSocket.PostSupported(const Handle: Hwnd): Boolean;
begin

	Result := true;
	try
  	PostHandle(Handle);
	except
		Result := false;
  end;

end;

procedure TMessageSocket.BroadcastIdentity;

var
	Handles: PHandles;

begin

	{La lista con gli handler dei componenti da cui in precedenza
   sono arrivate risposte  alla BroadcastIdentity va eliminata}
	while HandlesList <> nil do begin
		Handles := HandlesList;
    HandlesList := Handles^.Next;
    FreeMem(Handles);
  end;
  Windows.SendMessage(hwnd_Broadcast,WM_User + FLowMessageID,SendWindow,0);

end;

procedure TMessageSocket.SendMessage(const	MessageID: Word;
    																 const 	DestinationHandle: Hwnd;
    																 const 	DataPtr: Pointer;
                      							 const 	DataSize: Word);

var
  CopyData: TCopyDataStruct;

begin

	{Si invia un messaggio WM_Datacopy all'handle corrispondente ad un componente
   TMessageSocket, con WParam si trasferisce l'handle del componente da cui proviene il messaggio}
	if (MessageID > FLowMessageID) and (MessageID <= FHighMessageID) then begin
    CopyData.dwData := WM_User + MessageID;
    CopyData.cbData := DataSize;
    CopyData.lpData := DataPtr;
	  Windows.SendMessage(DestinationHandle,WM_COPYDATA,SendWindow,Longint(@CopyData));
  end
  else
  	raise EIllegalID.Create( TMessageSocket.ClassName + EIllegalIDMsg + IntToStr(MessageID))

end;

procedure TMessageSocket.PostMessage(const	MessageID: Word;
    																 const 	DestinationHandle: Hwnd;
    																 const 	DataPtr: Pointer;
                      							 const 	DataSize: Word);

var
  CopyData: TCopyDataStruct;

begin

	{Si invia un messaggio WM_Datacopy all'handle ottenuto con la PostHandle a
   partire dall'handle corrispondente ad un componente TMessageSocket, con
   WParam si trasferisce l'handle del componente da cui proviene il messaggio}
	if (MessageID > FLowMessageID) and (MessageID <= FHighMessageID) then begin
    CopyData.dwData := WM_User + MessageID;
    CopyData.cbData := DataSize;
    CopyData.lpData := DataPtr;
	  Windows.SendMessage(PostHandle(DestinationHandle),WM_COPYDATA,SendWindow,Longint(@CopyData));
  end
  else
  	raise EIllegalID.Create( TMessageSocket.ClassName + EIllegalIDMsg + IntToStr(MessageID))

end;

procedure TMessageSocket.MessageHandled;
begin

 	if Windows.InSendMessage then
   	Windows.ReplyMessage(Longint(True));

end;

{* Register ******************************************************************}

procedure Register;
begin

  RegisterComponents('Additional', [TMessageSocket]);

end;


end.

